/* -----------------------------------------

   _  _ ____ _    _    
   |__| |___ |    |    
   |  | |___ |___ |___ 
   
   HELL: High-Energy Large Logarithms

   Author: Marco Bonvini

   Computes Delta P_res = P_res - P_FixedOrder
   from a file containing parameters

   ----------------------------------------- */

#pragma once

#include <iostream>
#include <map>
#include "math/matrix.hh"
#include "hell_tables.hh"
#include "commons.hh"
#include "version.hh"

using namespace std;



#define gg entry11
#define gq entry12
#define qg entry21
#define qq entry22



namespace HELL {

  enum Order {
    none = -1,
    LO = 0,
    NLO = 1,
    NNLO = 2
  };
  enum LogOrder {
    LL = 0,
    NLL = 1
  };


  class Interpolator;
  class xTableU;
  class HiggsSmallX;
  extern void SetRCmodeCF   (unsigned int);
  extern void SetRCmodeCFold(unsigned int);
  extern void SetRCmodeCFnew(unsigned int);
  extern void SetRCvarCF   (bool);
  extern void SetRCvarCFold(bool);
  extern void SetLLpMode(bool use_LLp=true);

  // Class for fixed nf.
  //
  class HELLnf {
  private:
    int _nf;
    int _order;
    int _damping, _dampingsqrt;
    bool _momcons;
    bool _evaluateColdway;
    unsigned int _RCvar;
    unsigned int _RCmode;
    double _muFrat;
    vector<double> _alphas;
    string _datapath;
    map<int,HellTableP*>::iterator itHTP;
    map<int,HellTableP*> HTP;
    map<string,map<double,map<int,HellTableP*> > > HTP2;
    map<double,Interpolator*> Interp;
    map<double,Interpolator*>::iterator itInterp;
    map<double,HiggsSmallX*> ggHsx;
    map<double,HiggsSmallX*>::iterator itggHsx;
    map<int,xTableU*>::iterator itxTU;
    map<int,xTableU*> xTU[2][2];
    int alphas_interpolation(double as, vector<double> vas, double &factor);
    void ReadTable (int k);
    void ReadTableC(int k, string id, double d);
    void ReadTablex(int k);
    double xDeltaCRes(string id, double x, double as, double nf, xTableU *tU, void *pars=NULL);
    double xdeltaPplus(double as, double x);
    double xdeltaPqg  (double as, double x);
    double xdeltaC(double as, double x, string id, double mQ=0, double xi=0);
    sqmatrix<double>   DeltaPLL     (double x,  double as);
    sqmatrix<double>   DeltaPNLL    (double x,  double as);
    map<double,double> MC;
    map<double,double>::iterator itMC;
    friend double gMCintegrand(double,void*);
    double gMC(double as);
  public:
    HELLnf(int nf, LogOrder order, string datapath="./data/")
      : _nf(nf), _order(order), _damping(0), _dampingsqrt(0), _momcons(false), _evaluateColdway(true), _RCvar(false), _RCmode(0), _muFrat(0) //, ggHsx(NULL)
    { Init(datapath); }
    ~HELLnf();
    //
    void Init(string datapath);
    //
    int GetOrder() {return _order;};
    int GetNf() {return _nf;};
    void GetAvailableAlphas(vector<double> &as1);
    void SetEvaluationMethod(bool method=0) { _evaluateColdway = !method; } // set to 1 to use pretabulated U and Uqg tables, or 0 for standard usage
    void SetLargexDamping(int n=2, int m=4) { _damping=n; _dampingsqrt=m; } // multiplies deltaP by (1-x)^n (1-sqrt(x))^m
    void SetMomentumConservation(bool enable=true) { _momcons=enable; }
    void SetRCvariation(bool var=true) { _RCvar=(var?1:0); HELL::SetRCvarCF(var); HELL::SetRCvarCFold(var); } // using gamma+ with "Airy"
    void SetRCmode(unsigned int mode=1) { _RCmode=mode; HELL::SetRCmodeCF(mode); HELL::SetRCmodeCFold(mode); HELL::SetRCmodeCFnew(mode); } // r->as*b0
    void SetFactorizationScale(double muF_over_Q) { _muFrat = muF_over_Q; }
    //
    // Delta Splitting functions
    double deltaPplus(double as, double x);
    double deltaPLLp (double as, double x);
    double deltaPLLpd(double as, double x);
    double deltaP0   (double as, double x);
    double deltaP0sq (double as, double x);
    double deltaP0cb (double as, double x);
    double deltaP1   (double as, double x);
    double deltaP0P1 (double as, double x);
    double deltaP2   (double as, double x);
    double xdeltaPpUqg(double as, double x); // remove this
    sqmatrix<double> DeltaP(double as, double x);
    double deltaLogR(double as, double x);
    double deltaUqg (double as, double x);
    double deltaU   (double as, double x, double xi);
    //
    // Delta Coefficient functions (the quark coefficients are found multiplying by CF/CA)
    // these are delta2, to be matched to NNLO
    double deltaC2g   (double as, double x);  // DIS F2
    double deltaCLg   (double as, double x);  // DIS FL
    double deltaC2gold(double as, double x);  // DIS F2
    double deltaCLgold(double as, double x);  // DIS FL
    //
    double deltaKhg   (double as, double x, double m_Q_ratio);  // heavy quark matching condition
    // Neutral Current
    double deltaMC2g  (double as, double x, double m_Q_ratio);  // DIS F2 massive
    double deltaMCLg  (double as, double x, double m_Q_ratio);  // DIS FL massive
    double deltaMC2axg(double as, double x, double m_Q_ratio);  // DIS F2 massive, axial part, to be multiplied by gA^2/(gV^2+gA^2)
    double deltaMCLaxg(double as, double x, double m_Q_ratio);  // DIS FL massive, axial part, to be multiplied by gA^2/(gV^2+gA^2)
    // Charged Current
    double deltaMC2CCg(double as, double x, double m_Q_ratio);  // DIS F2 massive
    double deltaMCLCCg(double as, double x, double m_Q_ratio);  // DIS FL massive
    double deltaMC3CCg(double as, double x, double m_Q_ratio);  // DIS F3 massive
    //
    double deltaHgg   (double as, double x, double mH=125, double mt=173, double mb=-1, double mc=-1);
    double deltaHggAux(double as, double x, double mH=125, double mt=173, double mb=-1, double mc=-1);
    //
    double deltaCDYqg(double as, double x);
  };



};

